"""Generated wrapper for TWFactory Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for TWFactory below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        TWFactoryValidator,
    )
except ImportError:

    class TWFactoryValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass





class DefaultAdminRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the DEFAULT_ADMIN_ROLE method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method().call(tx_params.as_dict())
        return Union[bytes, str](returned)

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

class FactoryRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the FACTORY_ROLE method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method().call(tx_params.as_dict())
        return Union[bytes, str](returned)

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

class AddImplementationMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the addImplementation method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, implementation: str):
        """Validate the inputs to the addImplementation method."""
        self.validator.assert_valid(
            method_name='addImplementation',
            parameter_name='_implementation',
            argument_value=implementation,
        )
        implementation = self.validate_and_checksum_address(implementation)
        return (implementation)

    def call(self, implementation: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (implementation) = self.validate_and_normalize_inputs(implementation)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(implementation).call(tx_params.as_dict())

    def send_transaction(self, implementation: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (implementation) = self.validate_and_normalize_inputs(implementation)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation).transact(tx_params.as_dict())

    def build_transaction(self, implementation: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (implementation) = self.validate_and_normalize_inputs(implementation)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, implementation: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (implementation) = self.validate_and_normalize_inputs(implementation)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation).estimateGas(tx_params.as_dict())

class ApprovalMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the approval method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: str):
        """Validate the inputs to the approval method."""
        self.validator.assert_valid(
            method_name='approval',
            parameter_name='index_0',
            argument_value=index_0,
        )
        index_0 = self.validate_and_checksum_address(index_0)
        return (index_0)

    def call(self, index_0: str, tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(index_0).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(self, index_0: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).transact(tx_params.as_dict())

    def build_transaction(self, index_0: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, index_0: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).estimateGas(tx_params.as_dict())

class ApproveImplementationMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the approveImplementation method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, implementation: str, to_approve: bool):
        """Validate the inputs to the approveImplementation method."""
        self.validator.assert_valid(
            method_name='approveImplementation',
            parameter_name='_implementation',
            argument_value=implementation,
        )
        implementation = self.validate_and_checksum_address(implementation)
        self.validator.assert_valid(
            method_name='approveImplementation',
            parameter_name='_toApprove',
            argument_value=to_approve,
        )
        return (implementation, to_approve)

    def call(self, implementation: str, to_approve: bool, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (implementation, to_approve) = self.validate_and_normalize_inputs(implementation, to_approve)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(implementation, to_approve).call(tx_params.as_dict())

    def send_transaction(self, implementation: str, to_approve: bool, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (implementation, to_approve) = self.validate_and_normalize_inputs(implementation, to_approve)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation, to_approve).transact(tx_params.as_dict())

    def build_transaction(self, implementation: str, to_approve: bool, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (implementation, to_approve) = self.validate_and_normalize_inputs(implementation, to_approve)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation, to_approve).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, implementation: str, to_approve: bool, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (implementation, to_approve) = self.validate_and_normalize_inputs(implementation, to_approve)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation, to_approve).estimateGas(tx_params.as_dict())

class CurrentVersionMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the currentVersion method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: Union[bytes, str]):
        """Validate the inputs to the currentVersion method."""
        self.validator.assert_valid(
            method_name='currentVersion',
            parameter_name='index_0',
            argument_value=index_0,
        )
        return (index_0)

    def call(self, index_0: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(index_0).call(tx_params.as_dict())
        return int(returned)

    def send_transaction(self, index_0: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).transact(tx_params.as_dict())

    def build_transaction(self, index_0: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, index_0: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).estimateGas(tx_params.as_dict())

class DeployProxyMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the deployProxy method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _type: Union[bytes, str], data: Union[bytes, str]):
        """Validate the inputs to the deployProxy method."""
        self.validator.assert_valid(
            method_name='deployProxy',
            parameter_name='_type',
            argument_value=_type,
        )
        self.validator.assert_valid(
            method_name='deployProxy',
            parameter_name='_data',
            argument_value=data,
        )
        return (_type, data)

    def call(self, _type: Union[bytes, str], data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_type, data) = self.validate_and_normalize_inputs(_type, data)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(_type, data).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, _type: Union[bytes, str], data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (_type, data) = self.validate_and_normalize_inputs(_type, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, data).transact(tx_params.as_dict())

    def build_transaction(self, _type: Union[bytes, str], data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (_type, data) = self.validate_and_normalize_inputs(_type, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, data).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, _type: Union[bytes, str], data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (_type, data) = self.validate_and_normalize_inputs(_type, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, data).estimateGas(tx_params.as_dict())

class DeployProxyByImplementationMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the deployProxyByImplementation method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, implementation: str, data: Union[bytes, str], salt: Union[bytes, str]):
        """Validate the inputs to the deployProxyByImplementation method."""
        self.validator.assert_valid(
            method_name='deployProxyByImplementation',
            parameter_name='_implementation',
            argument_value=implementation,
        )
        implementation = self.validate_and_checksum_address(implementation)
        self.validator.assert_valid(
            method_name='deployProxyByImplementation',
            parameter_name='_data',
            argument_value=data,
        )
        self.validator.assert_valid(
            method_name='deployProxyByImplementation',
            parameter_name='_salt',
            argument_value=salt,
        )
        return (implementation, data, salt)

    def call(self, implementation: str, data: Union[bytes, str], salt: Union[bytes, str], tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (implementation, data, salt) = self.validate_and_normalize_inputs(implementation, data, salt)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(implementation, data, salt).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, implementation: str, data: Union[bytes, str], salt: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (implementation, data, salt) = self.validate_and_normalize_inputs(implementation, data, salt)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation, data, salt).transact(tx_params.as_dict())

    def build_transaction(self, implementation: str, data: Union[bytes, str], salt: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (implementation, data, salt) = self.validate_and_normalize_inputs(implementation, data, salt)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation, data, salt).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, implementation: str, data: Union[bytes, str], salt: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (implementation, data, salt) = self.validate_and_normalize_inputs(implementation, data, salt)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(implementation, data, salt).estimateGas(tx_params.as_dict())

class DeployProxyDeterministicMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the deployProxyDeterministic method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _type: Union[bytes, str], data: Union[bytes, str], salt: Union[bytes, str]):
        """Validate the inputs to the deployProxyDeterministic method."""
        self.validator.assert_valid(
            method_name='deployProxyDeterministic',
            parameter_name='_type',
            argument_value=_type,
        )
        self.validator.assert_valid(
            method_name='deployProxyDeterministic',
            parameter_name='_data',
            argument_value=data,
        )
        self.validator.assert_valid(
            method_name='deployProxyDeterministic',
            parameter_name='_salt',
            argument_value=salt,
        )
        return (_type, data, salt)

    def call(self, _type: Union[bytes, str], data: Union[bytes, str], salt: Union[bytes, str], tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_type, data, salt) = self.validate_and_normalize_inputs(_type, data, salt)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(_type, data, salt).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, _type: Union[bytes, str], data: Union[bytes, str], salt: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (_type, data, salt) = self.validate_and_normalize_inputs(_type, data, salt)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, data, salt).transact(tx_params.as_dict())

    def build_transaction(self, _type: Union[bytes, str], data: Union[bytes, str], salt: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (_type, data, salt) = self.validate_and_normalize_inputs(_type, data, salt)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, data, salt).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, _type: Union[bytes, str], data: Union[bytes, str], salt: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (_type, data, salt) = self.validate_and_normalize_inputs(_type, data, salt)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, data, salt).estimateGas(tx_params.as_dict())

class DeployerMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the deployer method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: str):
        """Validate the inputs to the deployer method."""
        self.validator.assert_valid(
            method_name='deployer',
            parameter_name='index_0',
            argument_value=index_0,
        )
        index_0 = self.validate_and_checksum_address(index_0)
        return (index_0)

    def call(self, index_0: str, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(index_0).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, index_0: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).transact(tx_params.as_dict())

    def build_transaction(self, index_0: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, index_0: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).estimateGas(tx_params.as_dict())

class GetImplementationMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getImplementation method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _type: Union[bytes, str], version: int):
        """Validate the inputs to the getImplementation method."""
        self.validator.assert_valid(
            method_name='getImplementation',
            parameter_name='_type',
            argument_value=_type,
        )
        self.validator.assert_valid(
            method_name='getImplementation',
            parameter_name='_version',
            argument_value=version,
        )
        # safeguard against fractional inputs
        version = int(version)
        return (_type, version)

    def call(self, _type: Union[bytes, str], version: int, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_type, version) = self.validate_and_normalize_inputs(_type, version)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(_type, version).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, _type: Union[bytes, str], version: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (_type, version) = self.validate_and_normalize_inputs(_type, version)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, version).transact(tx_params.as_dict())

    def build_transaction(self, _type: Union[bytes, str], version: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (_type, version) = self.validate_and_normalize_inputs(_type, version)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, version).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, _type: Union[bytes, str], version: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (_type, version) = self.validate_and_normalize_inputs(_type, version)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type, version).estimateGas(tx_params.as_dict())

class GetLatestImplementationMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getLatestImplementation method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _type: Union[bytes, str]):
        """Validate the inputs to the getLatestImplementation method."""
        self.validator.assert_valid(
            method_name='getLatestImplementation',
            parameter_name='_type',
            argument_value=_type,
        )
        return (_type)

    def call(self, _type: Union[bytes, str], tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_type) = self.validate_and_normalize_inputs(_type)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(_type).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, _type: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (_type) = self.validate_and_normalize_inputs(_type)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type).transact(tx_params.as_dict())

    def build_transaction(self, _type: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (_type) = self.validate_and_normalize_inputs(_type)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, _type: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (_type) = self.validate_and_normalize_inputs(_type)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_type).estimateGas(tx_params.as_dict())

class GetRoleAdminMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getRoleAdmin method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str]):
        """Validate the inputs to the getRoleAdmin method."""
        self.validator.assert_valid(
            method_name='getRoleAdmin',
            parameter_name='role',
            argument_value=role,
        )
        return (role)

    def call(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role).call(tx_params.as_dict())
        return Union[bytes, str](returned)

    def send_transaction(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).estimateGas(tx_params.as_dict())

class GetRoleMemberMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getRoleMember method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], index: int):
        """Validate the inputs to the getRoleMember method."""
        self.validator.assert_valid(
            method_name='getRoleMember',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='getRoleMember',
            parameter_name='index',
            argument_value=index,
        )
        # safeguard against fractional inputs
        index = int(index)
        return (role, index)

    def call(self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role, index).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).estimateGas(tx_params.as_dict())

class GetRoleMemberCountMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getRoleMemberCount method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str]):
        """Validate the inputs to the getRoleMemberCount method."""
        self.validator.assert_valid(
            method_name='getRoleMemberCount',
            parameter_name='role',
            argument_value=role,
        )
        return (role)

    def call(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role).call(tx_params.as_dict())
        return int(returned)

    def send_transaction(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).estimateGas(tx_params.as_dict())

class GrantRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the grantRole method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the grantRole method."""
        self.validator.assert_valid(
            method_name='grantRole',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='grantRole',
            parameter_name='account',
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())

class HasRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the hasRole method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the hasRole method."""
        self.validator.assert_valid(
            method_name='hasRole',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='hasRole',
            parameter_name='account',
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role, account).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())

class ImplementationMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the implementation method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: Union[bytes, str], index_1: int):
        """Validate the inputs to the implementation method."""
        self.validator.assert_valid(
            method_name='implementation',
            parameter_name='index_0',
            argument_value=index_0,
        )
        self.validator.assert_valid(
            method_name='implementation',
            parameter_name='index_1',
            argument_value=index_1,
        )
        # safeguard against fractional inputs
        index_1 = int(index_1)
        return (index_0, index_1)

    def call(self, index_0: Union[bytes, str], index_1: int, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (index_0, index_1) = self.validate_and_normalize_inputs(index_0, index_1)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(index_0, index_1).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, index_0: Union[bytes, str], index_1: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (index_0, index_1) = self.validate_and_normalize_inputs(index_0, index_1)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0, index_1).transact(tx_params.as_dict())

    def build_transaction(self, index_0: Union[bytes, str], index_1: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (index_0, index_1) = self.validate_and_normalize_inputs(index_0, index_1)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0, index_1).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, index_0: Union[bytes, str], index_1: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (index_0, index_1) = self.validate_and_normalize_inputs(index_0, index_1)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0, index_1).estimateGas(tx_params.as_dict())

class IsTrustedForwarderMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the isTrustedForwarder method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, forwarder: str):
        """Validate the inputs to the isTrustedForwarder method."""
        self.validator.assert_valid(
            method_name='isTrustedForwarder',
            parameter_name='forwarder',
            argument_value=forwarder,
        )
        forwarder = self.validate_and_checksum_address(forwarder)
        return (forwarder)

    def call(self, forwarder: str, tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (forwarder) = self.validate_and_normalize_inputs(forwarder)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(forwarder).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(self, forwarder: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (forwarder) = self.validate_and_normalize_inputs(forwarder)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(forwarder).transact(tx_params.as_dict())

    def build_transaction(self, forwarder: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (forwarder) = self.validate_and_normalize_inputs(forwarder)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(forwarder).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, forwarder: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (forwarder) = self.validate_and_normalize_inputs(forwarder)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(forwarder).estimateGas(tx_params.as_dict())

class MulticallMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the multicall method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, data: List[Union[bytes, str]]):
        """Validate the inputs to the multicall method."""
        self.validator.assert_valid(
            method_name='multicall',
            parameter_name='data',
            argument_value=data,
        )
        return (data)

    def call(self, data: List[Union[bytes, str]], tx_params: Optional[TxParams] = None) -> List[Union[bytes, str]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (data) = self.validate_and_normalize_inputs(data)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(data).call(tx_params.as_dict())
        return [Union[bytes, str](element) for element in returned]

    def send_transaction(self, data: List[Union[bytes, str]], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (data) = self.validate_and_normalize_inputs(data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(data).transact(tx_params.as_dict())

    def build_transaction(self, data: List[Union[bytes, str]], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (data) = self.validate_and_normalize_inputs(data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(data).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, data: List[Union[bytes, str]], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (data) = self.validate_and_normalize_inputs(data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(data).estimateGas(tx_params.as_dict())

class RegistryMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the registry method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method().call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

class RenounceRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the renounceRole method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the renounceRole method."""
        self.validator.assert_valid(
            method_name='renounceRole',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='renounceRole',
            parameter_name='account',
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())

class RevokeRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the revokeRole method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the revokeRole method."""
        self.validator.assert_valid(
            method_name='revokeRole',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='revokeRole',
            parameter_name='account',
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())

class SupportsInterfaceMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the supportsInterface method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, interface_id: Union[bytes, str]):
        """Validate the inputs to the supportsInterface method."""
        self.validator.assert_valid(
            method_name='supportsInterface',
            parameter_name='interfaceId',
            argument_value=interface_id,
        )
        return (interface_id)

    def call(self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(interface_id).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).transact(tx_params.as_dict())

    def build_transaction(self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).estimateGas(tx_params.as_dict())

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class TWFactory:
    """Wrapper class for TWFactory Solidity contract.

    All method parameters of type `bytes`:code: should be encoded as UTF-8,
    which can be accomplished via `str.encode("utf_8")`:code:.
    """
    default_admin_role: DefaultAdminRoleMethod
    """Constructor-initialized instance of
    :class:`DefaultAdminRoleMethod`.
    """

    factory_role: FactoryRoleMethod
    """Constructor-initialized instance of
    :class:`FactoryRoleMethod`.
    """

    add_implementation: AddImplementationMethod
    """Constructor-initialized instance of
    :class:`AddImplementationMethod`.
    """

    approval: ApprovalMethod
    """Constructor-initialized instance of
    :class:`ApprovalMethod`.
    """

    approve_implementation: ApproveImplementationMethod
    """Constructor-initialized instance of
    :class:`ApproveImplementationMethod`.
    """

    current_version: CurrentVersionMethod
    """Constructor-initialized instance of
    :class:`CurrentVersionMethod`.
    """

    deploy_proxy: DeployProxyMethod
    """Constructor-initialized instance of
    :class:`DeployProxyMethod`.
    """

    deploy_proxy_by_implementation: DeployProxyByImplementationMethod
    """Constructor-initialized instance of
    :class:`DeployProxyByImplementationMethod`.
    """

    deploy_proxy_deterministic: DeployProxyDeterministicMethod
    """Constructor-initialized instance of
    :class:`DeployProxyDeterministicMethod`.
    """

    deployer: DeployerMethod
    """Constructor-initialized instance of
    :class:`DeployerMethod`.
    """

    get_implementation: GetImplementationMethod
    """Constructor-initialized instance of
    :class:`GetImplementationMethod`.
    """

    get_latest_implementation: GetLatestImplementationMethod
    """Constructor-initialized instance of
    :class:`GetLatestImplementationMethod`.
    """

    get_role_admin: GetRoleAdminMethod
    """Constructor-initialized instance of
    :class:`GetRoleAdminMethod`.
    """

    get_role_member: GetRoleMemberMethod
    """Constructor-initialized instance of
    :class:`GetRoleMemberMethod`.
    """

    get_role_member_count: GetRoleMemberCountMethod
    """Constructor-initialized instance of
    :class:`GetRoleMemberCountMethod`.
    """

    grant_role: GrantRoleMethod
    """Constructor-initialized instance of
    :class:`GrantRoleMethod`.
    """

    has_role: HasRoleMethod
    """Constructor-initialized instance of
    :class:`HasRoleMethod`.
    """

    implementation: ImplementationMethod
    """Constructor-initialized instance of
    :class:`ImplementationMethod`.
    """

    is_trusted_forwarder: IsTrustedForwarderMethod
    """Constructor-initialized instance of
    :class:`IsTrustedForwarderMethod`.
    """

    multicall: MulticallMethod
    """Constructor-initialized instance of
    :class:`MulticallMethod`.
    """

    registry: RegistryMethod
    """Constructor-initialized instance of
    :class:`RegistryMethod`.
    """

    renounce_role: RenounceRoleMethod
    """Constructor-initialized instance of
    :class:`RenounceRoleMethod`.
    """

    revoke_role: RevokeRoleMethod
    """Constructor-initialized instance of
    :class:`RevokeRoleMethod`.
    """

    supports_interface: SupportsInterfaceMethod
    """Constructor-initialized instance of
    :class:`SupportsInterfaceMethod`.
    """


    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: TWFactoryValidator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = TWFactoryValidator(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        else:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                         middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=TWFactory.abi()).functions

        self.default_admin_role = DefaultAdminRoleMethod(web3_or_provider, contract_address, functions.DEFAULT_ADMIN_ROLE)

        self.factory_role = FactoryRoleMethod(web3_or_provider, contract_address, functions.FACTORY_ROLE)

        self.add_implementation = AddImplementationMethod(web3_or_provider, contract_address, functions.addImplementation, validator)

        self.approval = ApprovalMethod(web3_or_provider, contract_address, functions.approval, validator)

        self.approve_implementation = ApproveImplementationMethod(web3_or_provider, contract_address, functions.approveImplementation, validator)

        self.current_version = CurrentVersionMethod(web3_or_provider, contract_address, functions.currentVersion, validator)

        self.deploy_proxy = DeployProxyMethod(web3_or_provider, contract_address, functions.deployProxy, validator)

        self.deploy_proxy_by_implementation = DeployProxyByImplementationMethod(web3_or_provider, contract_address, functions.deployProxyByImplementation, validator)

        self.deploy_proxy_deterministic = DeployProxyDeterministicMethod(web3_or_provider, contract_address, functions.deployProxyDeterministic, validator)

        self.deployer = DeployerMethod(web3_or_provider, contract_address, functions.deployer, validator)

        self.get_implementation = GetImplementationMethod(web3_or_provider, contract_address, functions.getImplementation, validator)

        self.get_latest_implementation = GetLatestImplementationMethod(web3_or_provider, contract_address, functions.getLatestImplementation, validator)

        self.get_role_admin = GetRoleAdminMethod(web3_or_provider, contract_address, functions.getRoleAdmin, validator)

        self.get_role_member = GetRoleMemberMethod(web3_or_provider, contract_address, functions.getRoleMember, validator)

        self.get_role_member_count = GetRoleMemberCountMethod(web3_or_provider, contract_address, functions.getRoleMemberCount, validator)

        self.grant_role = GrantRoleMethod(web3_or_provider, contract_address, functions.grantRole, validator)

        self.has_role = HasRoleMethod(web3_or_provider, contract_address, functions.hasRole, validator)

        self.implementation = ImplementationMethod(web3_or_provider, contract_address, functions.implementation, validator)

        self.is_trusted_forwarder = IsTrustedForwarderMethod(web3_or_provider, contract_address, functions.isTrustedForwarder, validator)

        self.multicall = MulticallMethod(web3_or_provider, contract_address, functions.multicall, validator)

        self.registry = RegistryMethod(web3_or_provider, contract_address, functions.registry)

        self.renounce_role = RenounceRoleMethod(web3_or_provider, contract_address, functions.renounceRole, validator)

        self.revoke_role = RevokeRoleMethod(web3_or_provider, contract_address, functions.revokeRole, validator)

        self.supports_interface = SupportsInterfaceMethod(web3_or_provider, contract_address, functions.supportsInterface, validator)

    def get_implementation_added_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for ImplementationAdded event.

        :param tx_hash: hash of transaction emitting ImplementationAdded event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=TWFactory.abi()).events.ImplementationAdded().processReceipt(tx_receipt)
    def get_implementation_approved_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for ImplementationApproved event.

        :param tx_hash: hash of transaction emitting ImplementationApproved
            event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=TWFactory.abi()).events.ImplementationApproved().processReceipt(tx_receipt)
    def get_proxy_deployed_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for ProxyDeployed event.

        :param tx_hash: hash of transaction emitting ProxyDeployed event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=TWFactory.abi()).events.ProxyDeployed().processReceipt(tx_receipt)
    def get_role_admin_changed_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleAdminChanged event.

        :param tx_hash: hash of transaction emitting RoleAdminChanged event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=TWFactory.abi()).events.RoleAdminChanged().processReceipt(tx_receipt)
    def get_role_granted_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleGranted event.

        :param tx_hash: hash of transaction emitting RoleGranted event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=TWFactory.abi()).events.RoleGranted().processReceipt(tx_receipt)
    def get_role_revoked_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleRevoked event.

        :param tx_hash: hash of transaction emitting RoleRevoked event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=TWFactory.abi()).events.RoleRevoked().processReceipt(tx_receipt)

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"inputs":[{"internalType":"address","name":"_trustedForwarder","type":"address"},{"internalType":"address","name":"_registry","type":"address"}],"stateMutability":"nonpayable","type":"constructor"},{"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"implementation","type":"address"},{"indexed":true,"internalType":"bytes32","name":"contractType","type":"bytes32"},{"indexed":false,"internalType":"uint256","name":"version","type":"uint256"}],"name":"ImplementationAdded","type":"event"},{"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"implementation","type":"address"},{"indexed":false,"internalType":"bool","name":"isApproved","type":"bool"}],"name":"ImplementationApproved","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"implementation","type":"address"},{"indexed":false,"internalType":"address","name":"proxy","type":"address"},{"indexed":true,"internalType":"address","name":"deployer","type":"address"}],"name":"ProxyDeployed","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"previousAdminRole","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"newAdminRole","type":"bytes32"}],"name":"RoleAdminChanged","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleGranted","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleRevoked","type":"event"},{"inputs":[],"name":"DEFAULT_ADMIN_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"FACTORY_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_implementation","type":"address"}],"name":"addImplementation","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"index_0","type":"address"}],"name":"approval","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_implementation","type":"address"},{"internalType":"bool","name":"_toApprove","type":"bool"}],"name":"approveImplementation","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"index_0","type":"bytes32"}],"name":"currentVersion","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_type","type":"bytes32"},{"internalType":"bytes","name":"_data","type":"bytes"}],"name":"deployProxy","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_implementation","type":"address"},{"internalType":"bytes","name":"_data","type":"bytes"},{"internalType":"bytes32","name":"_salt","type":"bytes32"}],"name":"deployProxyByImplementation","outputs":[{"internalType":"address","name":"deployedProxy","type":"address"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_type","type":"bytes32"},{"internalType":"bytes","name":"_data","type":"bytes"},{"internalType":"bytes32","name":"_salt","type":"bytes32"}],"name":"deployProxyDeterministic","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"index_0","type":"address"}],"name":"deployer","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_type","type":"bytes32"},{"internalType":"uint256","name":"_version","type":"uint256"}],"name":"getImplementation","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_type","type":"bytes32"}],"name":"getLatestImplementation","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleAdmin","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"uint256","name":"index","type":"uint256"}],"name":"getRoleMember","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleMemberCount","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"grantRole","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"hasRole","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"index_0","type":"bytes32"},{"internalType":"uint256","name":"index_1","type":"uint256"}],"name":"implementation","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"forwarder","type":"address"}],"name":"isTrustedForwarder","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes[]","name":"data","type":"bytes[]"}],"name":"multicall","outputs":[{"internalType":"bytes[]","name":"results","type":"bytes[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"registry","outputs":[{"internalType":"contract TWRegistry","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"renounceRole","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"revokeRole","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes4","name":"interfaceId","type":"bytes4"}],"name":"supportsInterface","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"}]'  # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



from typing import Any, Callable, Dict, Generic
from thirdweb.constants.events import EventType
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.types.contract import TContractABI
from thirdweb.types.events import TxEvent


class ContractEvents(Generic[TContractABI]):
    _contract_wrapper: ContractWrapper[TContractABI]

    def __init__(self, contract_wrapper: ContractWrapper[TContractABI]):
        self._contract_wrapper = contract_wrapper

    def add_transaction_listener(self, listener: Callable[[TxEvent], Any]):
        """
        Add a listener for transaction events to this contract. This will be called
        whenever a transaction event is executed from the SDK interface.

        :param listener: The listener function to be called on transaction events.
        """

        self._contract_wrapper.add_listener(EventType.TRANSACTION, listener)  # type: ignore

    def remove_transaction_listener(self, listener: Callable[[TxEvent], Any]):
        """
        Remove a listener from transaction events.

        :param listener: The listener function to be removed.
        """

        self._contract_wrapper.remove_listener(EventType.TRANSACTION, listener)  # type: ignore

    def add_event_listener(
        self, event_name: str, listener: Callable[[Dict[str, Any]], None]
    ):
        """
        Add an event listener to this contract to listen for a specific event type.

        :param event_name: The name of the event to listen for.
        :param listener: The listener function to be called on the event.
        """

        interface = self._contract_wrapper.get_contract_interface()
        event = interface.events[event_name].abi
        event_names = [i["name"] for i in event["inputs"]]

        def event_listener(*args):
            results: Dict[str, Any] = {}
            for i, arg in enumerate(event_names):
                results[arg] = args[i]
            listener(results)

        self._contract_wrapper.on(event["name"], event_listener)

    def remove_event_listener(self, event_name: str, listener):
        """
        Remove an event listener listening to a specific event type.

        :param event_name: The name of the event to stop listening for.
        :param listener: The listener function to be removed.
        """

        interface = self._contract_wrapper.get_contract_interface()
        event = interface.events[event_name].abi

        self._contract_wrapper.remove_listener(event["name"], listener)

    def remove_all_listeners(self):
        """
        Remove all event listeners from this contract.
        """

        self._contract_wrapper.remove_all_listeners()



from typing import Generic
from thirdweb.core.classes.contract_metadata import ContractMetadata
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.types.contract import TRoyaltyABI
from thirdweb.types.settings.metadata import ContractRoyaltySchema
from web3.eth import TxReceipt


class ContractRoyalty(Generic[TRoyaltyABI]):
    _contract_wrapper: ContractWrapper[TRoyaltyABI]
    _metadata: ContractMetadata

    def __init__(
        self, contract_wrapper: ContractWrapper[TRoyaltyABI], metadata: ContractMetadata
    ):
        self._contract_wrapper = contract_wrapper
        self._metadata = metadata

    def get_default_royalty_info(self) -> ContractRoyaltySchema:
        """
        Get the default royalty information for this contract.

        :returns: the default royalty information.
        """

        (
            royalty_recipient,
            royalty_bps,
        ) = self._contract_wrapper._contract_abi.get_default_royalty_info.call()
        return ContractRoyaltySchema(royalty_recipient, royalty_bps)

    def get_token_royalty_info(self, token_id: int) -> ContractRoyaltySchema:
        """
        Get the royalty information for a specific token.

        :param token_id: the id of the token.
        :returns: the royalty information for the token.
        """

        (
            royalty_recipient,
            royalty_bps,
        ) = self._contract_wrapper._contract_abi.get_royalty_info_for_token.call(
            token_id
        )
        return ContractRoyaltySchema(royalty_recipient, royalty_bps)

    def set_default_royalty_info(
        self, royalty_data: ContractRoyaltySchema
    ) -> TxReceipt:
        """
        Set the default royalty information for this contract.

        :param royalty_data: the default royalty information.
        :returns: the transaction receipt of setting the royalty.
        """

        metadata = self._metadata.get()
        metadata.fee_recipient = royalty_data.fee_recipient
        metadata.seller_fee_basis_points = royalty_data.seller_fee_basis_points

        contract_uri = self._metadata._parse_and_upload_metadata(metadata)

        encoded = [
            self._contract_wrapper.get_contract_interface().encodeABI(
                "set_default_royalty_info",
                [metadata.fee_recipient, metadata.seller_fee_basis_points],
            ),
            self._contract_wrapper.get_contract_interface().encodeABI(
                "set_contract_uri", [contract_uri]
            ),
        ]

        return self._contract_wrapper.multi_call(encoded)

    def set_token_royalty_info(
        self, token_id: int, royalty_data: ContractRoyaltySchema
    ) -> TxReceipt:
        """
        Set the royalty information for a specific token.

        :param token_id: the id of the token.
        :param royalty_data: the royalty information for the token.
        :returns: the transaction receipt of setting the royalty.
        """

        return self._contract_wrapper.send_transaction(
            "set_royalty_info_for_token",
            [
                token_id,
                royalty_data.fee_recipient,
                royalty_data.seller_fee_basis_points,
            ],
        )



from typing import Generic
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.types.contract import TPlatformFeeABI
from thirdweb.types.settings.metadata import ContractPlatformFeeSchema
from web3.eth import TxReceipt


class ContractPlatformFee(Generic[TPlatformFeeABI]):
    _contract_wrapper: ContractWrapper[TPlatformFeeABI]

    def __init__(self, contract_wrapper: ContractWrapper[TPlatformFeeABI]):
        self._contract_wrapper = contract_wrapper

    def get(self) -> ContractPlatformFeeSchema:
        """
        Get the platform fee of this contract.

        :returns: the platform fee.
        """

        (
            platform_fee_recipient,
            platform_fee_bps,
        ) = self._contract_wrapper._contract_abi.get_platform_fee_info.call()
        return ContractPlatformFeeSchema(platform_fee_recipient, platform_fee_bps)

    def set(self, platform_fee_info: ContractPlatformFeeSchema) -> TxReceipt:
        """
        Set the platform fee of this contract.

        :param platform_fee_info: the platform fee info to set.
        :returns: the transaction receipt of setting the platform fee.
        """

        return self._contract_wrapper.send_transaction(
            "set_platform_fee_info",
            [
                platform_fee_info.platform_fee_recipient,
                platform_fee_info.platform_fee_basis_points,
            ],
        )



from typing import Dict
from thirdweb.constants.chains import ChainId
from thirdweb.constants.currency import ZERO_ADDRESS

DEFAULT_MERKLE_ROOT = (
    "0x0000000000000000000000000000000000000000000000000000000000000000"
)

OZ_DEFENDER_FORWARDER_ADDRESS = "0xc82BbE41f2cF04e3a8efA18F7032BDD7f6d98a81"
TWREGISTRY_ADDRESS = "0x7c487845f98938Bb955B1D5AD069d9a30e4131fd"
TWFACTORY_ADDRESS = "0x11c34F062Cb10a20B9F463E12Ff9dA62D76FDf65"

CONTRACT_ADDRESSES: Dict[ChainId, Dict[str, str]] = {
    ChainId.MAINNET: {
        "biconomy_forwarder": "0x84a0856b038eaAd1cC7E297cF34A7e72685A8693",
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
    ChainId.RINKEBY: {
        "biconomy_forwarder": "0xFD4973FeB2031D4409fB57afEE5dF2051b171104",
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
    ChainId.GOERLI: {
        "biconomy_forwarder": ZERO_ADDRESS,
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
    ChainId.POLYGON: {
        "biconomy_forwarder": "0x86C80a8aa58e0A4fa09A69624c31Ab2a6CAD56b8",
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
    ChainId.MUMBAI: {
        "biconomy_forwarder": "0x9399BB24DBB5C4b782C70c2969F58716Ebbd6a3b",
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
    ChainId.AVALANCHE: {
        "biconomy_forwarder": "0x64CD353384109423a966dCd3Aa30D884C9b2E057",
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
    ChainId.FANTOM: {
        "biconomy_forwarder": ZERO_ADDRESS,
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
    ChainId.LOCALHOST: {
        "biconomy_forwarder": ZERO_ADDRESS,
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
    ChainId.HARDHAT: {
        "biconomy_forwarder": ZERO_ADDRESS,
        "tw_factory": TWFACTORY_ADDRESS,
        "tw_registry": TWREGISTRY_ADDRESS,
    },
}


def get_contract_address_by_chain_id(chain_id: ChainId, contract_name: str) -> str:
    return CONTRACT_ADDRESSES[chain_id][contract_name]



from typing import Any, Dict, List, Optional, cast
from thirdweb.abi.drop_erc721 import DropERC721
from thirdweb.abi.ierc20 import IERC20
from thirdweb.common.claim_conditions import (
    get_claimer_proofs,
    process_claim_condition_inputs,
    transform_result_to_claim_condition,
)
from thirdweb.common.currency import is_native_token, parse_units
from thirdweb.common.error import includes_error_message
from thirdweb.constants.addresses import DEFAULT_MERKLE_ROOT
from thirdweb.core.classes.contract_metadata import ContractMetadata
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.types.contracts.claim_conditions import (
    ClaimCondition,
    ClaimConditionInput,
    ClaimEligibility,
)
from thirdweb.types.currency import Amount
from thirdweb.types.settings.metadata import NFTDropContractMetadata
from web3.eth import TxReceipt
from web3.constants import MAX_INT
from time import time


class DropClaimConditions:
    _contract_wrapper: ContractWrapper[DropERC721]
    _metadata: ContractMetadata[DropERC721, NFTDropContractMetadata]
    _storage: IpfsStorage

    def __init__(
        self,
        contract_wrapper: ContractWrapper[DropERC721],
        metadata: ContractMetadata[DropERC721, NFTDropContractMetadata],
        storage: IpfsStorage,
    ):
        self._contract_wrapper = contract_wrapper
        self._metadata = metadata
        self._storage = storage

    """
    READ FUNCTIONS
    """

    def get_active(self) -> ClaimCondition:
        """
        Get the currently active claim condition

        :return: The currently active claim condition
        """

        id = self._contract_wrapper._contract_abi.get_active_claim_condition_id.call()
        mc = self._contract_wrapper._contract_abi.get_claim_condition_by_id.call(id)
        metadata = self._metadata.get()
        return transform_result_to_claim_condition(
            mc,
            self._contract_wrapper.get_provider(),
            metadata.merkle,
            self._storage,
        )

    def get_all(self) -> List[ClaimCondition]:
        """
        Get all claim conditions on this contract

        :return: A list of all claim conditions on this contract
        """

        # Is this the correct order?
        start_id, count = self._contract_wrapper._contract_abi.claim_condition.call()
        conditions = []
        for i in range(start_id + count):
            conditions.append(
                self._contract_wrapper._contract_abi.get_claim_condition_by_id.call(i)
            )
        metadata = self._metadata.get()
        return [
            transform_result_to_claim_condition(
                c, self._contract_wrapper.get_provider(), metadata.merkle, self._storage
            )
            for c in conditions
        ]

    # def can_claim(self, quantity: int, address_to_check: Optional[str] = None) -> bool:
    #     """
    #     Check if a specified wallet can claim a specified quantity of NFTs

    #     :param quantity: The quantity of NFTs to check
    #     :param address_to_check: The wallet to check
    #     :return: True if the wallet can claim the quantity of NFTs, False otherwise
    #     """

    #     address_to_check = (
    #         address_to_check
    #         if address_to_check is not None
    #         else self._contract_wrapper.get_signer_address()
    #     )

    #     print(
    #         "REASONS: ",
    #         self.get_claim_ineligibility_reasons(quantity, address_to_check),
    #     )

    #     return (
    #         len(self.get_claim_ineligibility_reasons(quantity, address_to_check)) == 0
    #     )

    # def get_claim_ineligibility_reasons(
    #     self, quantity: int, address_to_check: Optional[str] = None
    # ) -> List[ClaimEligibility]:
    #     """
    #     Get the reasons why a wallet cannot claim a specified quantity of NFTs

    #     :param quantity: The quantity of NFTs to check
    #     :param address_to_check: The wallet to check
    #     :return: A list of reasons why the wallet cannot claim the quantity of NFTs
    #     """

    #     reasons: List[ClaimEligibility] = []
    #     quantity_with_decimals = parse_units(quantity, self._get_token_decimals())

    #     if address_to_check == None:
    #         address_to_check = self._contract_wrapper.get_signer_address()

    #     try:
    #         active_condition_index = (
    #             self._contract_wrapper._contract_abi.get_active_claim_condition_id.call()
    #         )
    #         claim_condition = self.get_active()
    #     except Exception as e:
    #         if includes_error_message(e, "no public mint condition."):
    #             reasons.append(ClaimEligibility.NO_CLAIM_CONDITION_SET)
    #             return reasons
    #         if includes_error_message(e, "no active mint condition."):
    #             reasons.append(ClaimEligibility.NO_ACTIVE_CLAIM_PHASE)
    #             return reasons
    #         reasons.append(ClaimEligibility.UNKNOWN)
    #         return reasons

    #     if claim_condition.available_supply < quantity_with_decimals:
    #         reasons.append(ClaimEligibility.NOT_ENOUGH_SUPPLY)

    #     if (
    #         claim_condition.merkle_root_hash != DEFAULT_MERKLE_ROOT
    #         and claim_condition.merkle_root_hash
    #         != b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    #     ):
    #         merkle_lower = claim_condition.merkle_root_hash
    #         metadata = self._metadata.get()
    #         proofs = get_claimer_proofs(
    #             cast(str, address_to_check),
    #             merkle_lower,
    #             self._get_token_decimals(),
    #             metadata.merkle,
    #             self._storage,
    #         )

    #         try:
    #             (
    #                 valid_merkle_proof,
    #             ) = self._contract_wrapper._contract_abi.verify_claim_merkle_proof.call(
    #                 active_condition_index,
    #                 cast(str, address_to_check),
    #                 quantity,
    #                 proofs.proof,  # type: ignore
    #                 proofs.max_claimable,
    #             )

    #             if not valid_merkle_proof:  # type: ignore
    #                 reasons.append(ClaimEligibility.ADDRESS_NOT_ALLOWED)
    #                 return reasons
    #         except:
    #             reasons.append(ClaimEligibility.ADDRESS_NOT_ALLOWED)
    #             return reasons

    #     (
    #         last_claimed_timestamp,
    #         timestamp_for_next_claim,
    #     ) = self._contract_wrapper._contract_abi.get_claim_timestamp.call(
    #         active_condition_index, cast(str, address_to_check)
    #     )

    #     now = int(time())

    #     if last_claimed_timestamp < 0 and now < timestamp_for_next_claim:
    #         if timestamp_for_next_claim == int(MAX_INT, 0):
    #             reasons.append(ClaimEligibility.ALREADY_CLAIMED)
    #         else:
    #             reasons.append(ClaimEligibility.WAIT_BEFORE_NEXT_CLAIM_TRANSACTION)

    #     if claim_condition.price > 0:
    #         total_price = claim_condition.price * quantity
    #         provider = self._contract_wrapper.get_provider()
    #         if is_native_token(claim_condition.currency_address):
    #             balance = int(provider.eth.get_balance(address_to_check))
    #             if balance < total_price:
    #                 reasons.append(ClaimEligibility.NOT_ENOUGH_TOKENS)
    #         else:
    #             abi = IERC20(provider, claim_condition.currency_address)
    #             erc20 = ContractWrapper[IERC20](
    #                 abi, provider, self._contract_wrapper.get_signer()
    #             )
    #             balance = erc20._contract_abi.balance_of.call(
    #                 cast(str, address_to_check)
    #             )
    #             if balance < total_price:
    #                 reasons.append(ClaimEligibility.NOT_ENOUGH_TOKENS)

    #     return reasons

    """
    WRITE FUNCTIONS
    """

    # def set(
    #     self,
    #     claim_condition_inputs: List[ClaimConditionInput],
    #     reset_claim_eligibility_for_all=False,
    # ) -> TxReceipt:
    #     """
    #     Set the claim conditions for this contract

    #     :param claim_condition_inputs: List of claim condition inputs
    #     :param reset_claim_eligibility_for_all: Reset claim eligibility for all wallets
    #     :return: transaction receipt of the set
    #     """

    #     snapshot_infos, sorted_conditions = process_claim_condition_inputs(
    #         claim_condition_inputs,
    #         self._get_token_decimals(),
    #         self._contract_wrapper.get_provider(),
    #         self._storage,
    #     )

    #     merkle_info: Dict[str, Any] = {}
    #     for s in snapshot_infos:
    #         merkle_info[s.merkle_root] = s.snapshot_uri

    #     metadata = self._metadata.get()
    #     encoded = []
    #     interface = self._contract_wrapper.get_contract_interface()

    #     if not metadata.merkle == merkle_info:
    #         metadata.merkle = merkle_info
    #         contract_uri = self._metadata._parse_and_upload_metadata(metadata.to_json())
    #         encoded.append(interface.encodeABI("setContractURI", [contract_uri]))

    #     print("CONDITIONS ", sorted_conditions)

    #     encoded.append(
    #         interface.encodeABI(
    #             "setClaimConditions",
    #             [sorted_conditions, reset_claim_eligibility_for_all],
    #         )
    #     )

    #     return self._contract_wrapper.multi_call(encoded)

    """
    INTERNAL FUNCTIONS
    """

    def _get_token_decimals(self) -> int:
        # TODO: Implement for TokenERC20
        return 0



"""Generated wrapper for ThirdwebContract Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for ThirdwebContract below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        ThirdwebContractValidator,
    )
except ImportError:

    class ThirdwebContractValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass





class OwnerMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the owner method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method().call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

class SetOwnerMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the setOwner method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, new_owner: str):
        """Validate the inputs to the setOwner method."""
        self.validator.assert_valid(
            method_name='setOwner',
            parameter_name='_newOwner',
            argument_value=new_owner,
        )
        new_owner = self.validate_and_checksum_address(new_owner)
        return (new_owner)

    def call(self, new_owner: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(new_owner).call(tx_params.as_dict())

    def send_transaction(self, new_owner: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(new_owner).transact(tx_params.as_dict())

    def build_transaction(self, new_owner: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(new_owner).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, new_owner: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(new_owner).estimateGas(tx_params.as_dict())

class TwInitializeOwnerMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the tw_initializeOwner method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, deployer: str):
        """Validate the inputs to the tw_initializeOwner method."""
        self.validator.assert_valid(
            method_name='tw_initializeOwner',
            parameter_name='deployer',
            argument_value=deployer,
        )
        deployer = self.validate_and_checksum_address(deployer)
        return (deployer)

    def call(self, deployer: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(deployer).call(tx_params.as_dict())

    def send_transaction(self, deployer: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).transact(tx_params.as_dict())

    def build_transaction(self, deployer: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, deployer: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).estimateGas(tx_params.as_dict())

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class ThirdwebContract:
    """Wrapper class for ThirdwebContract Solidity contract."""
    owner: OwnerMethod
    """Constructor-initialized instance of
    :class:`OwnerMethod`.
    """

    set_owner: SetOwnerMethod
    """Constructor-initialized instance of
    :class:`SetOwnerMethod`.
    """

    tw_initialize_owner: TwInitializeOwnerMethod
    """Constructor-initialized instance of
    :class:`TwInitializeOwnerMethod`.
    """


    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: ThirdwebContractValidator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = ThirdwebContractValidator(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        else:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                         middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=ThirdwebContract.abi()).functions

        self.owner = OwnerMethod(web3_or_provider, contract_address, functions.owner)

        self.set_owner = SetOwnerMethod(web3_or_provider, contract_address, functions.setOwner, validator)

        self.tw_initialize_owner = TwInitializeOwnerMethod(web3_or_provider, contract_address, functions.tw_initializeOwner, validator)

    def get_owner_updated_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for OwnerUpdated event.

        :param tx_hash: hash of transaction emitting OwnerUpdated event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=ThirdwebContract.abi()).events.OwnerUpdated().processReceipt(tx_receipt)

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"prevOwner","type":"address"},{"indexed":false,"internalType":"address","name":"newOwner","type":"address"}],"name":"OwnerUpdated","type":"event"},{"inputs":[],"name":"owner","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_newOwner","type":"address"}],"name":"setOwner","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"deployer","type":"address"}],"name":"tw_initializeOwner","outputs":[],"stateMutability":"nonpayable","type":"function"}]'  # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



from typing import Dict, List
from thirdweb.common.error import RoleException
from thirdweb.constants.role import Role, get_role_hash
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from web3.eth import TxReceipt


class ContractRoles:
    _contract_wrapper: ContractWrapper
    _roles: List[Role]

    def __init__(self, contract_wrapper: ContractWrapper, roles: List[Role]):
        self._contract_wrapper = contract_wrapper
        self._roles = roles

    """
    READ FUNCTIONS
    """

    def get_all(self) -> Dict[Role, List[str]]:
        """
        Get all role members on this contract.

        :returns: a dictionary of role members for each role
        """

        if len(self._roles) == 0:
            raise RoleException("This contract has no support for roles")

        roles = {}
        for role in self._roles:
            roles[role] = self.get(role)

        return roles

    def get(self, role: Role) -> List[str]:
        """
        Get all members of a role on this contract.

        :param role: role to get members of
        :returns: list of members of the role
        """

        if role not in self._roles:
            raise RoleException(f"This contract does not support role {role}")

        role_hash = get_role_hash(role)
        count = self._contract_wrapper._contract_abi.get_role_member_count.call(
            role_hash
        )

        return [
            self._contract_wrapper._contract_abi.get_role_member.call(role_hash, i)
            for i in range(count)
        ]

    def verify(self, roles: List[Role], address: str):
        for role in roles:
            members = self.get(role)
            if address.lower() not in [member.lower() for member in members]:
                raise RoleException(f"Address {address} is not a member of role {role}")

    """
    WRITE FUNCTIONS
    """

    def grant(self, role: Role, address: str) -> TxReceipt:
        """
        Grant a role to an address.

        :param role: role to grant
        :param address: address to grant the role to
        :returns: transaction receipt of granting the role
        """

        if role not in self._roles:
            raise RoleException(f"This contract does not support role {role}")

        role_hash = get_role_hash(role)
        return self._contract_wrapper.send_transaction(
            "grant_role", [role_hash, address]
        )

    def revoke(self, role: Role, address: str) -> TxReceipt:
        """
        Revoke a role from an address.

        :param role: role to revoke
        :param address: address to revoke the role from
        :returns: transaction receipt of revoking the role
        """

        if role not in self._roles:
            raise RoleException(f"This contract does not support role {role}")

        role_hash = get_role_hash(role)
        revoke_function_name = self.get_revoke_role_function_name(address)
        return self._contract_wrapper.send_transaction(
            revoke_function_name, [role_hash, address]
        )

    def get_revoke_role_function_name(self, address: str) -> str:
        signer_address = self._contract_wrapper.get_signer_address()
        if signer_address.lower() == address.lower():
            return "renounce_role"
        return "revoke_role"



from typing import Any, Dict, Union

# from thirdweb.abi.token_erc1155.bytecode import bytecode as TOKEN_ERC1155_BYTECODE
# from thirdweb.abi.token_erc20.bytecode import bytecode as TOKEN_ERC20_BYTECODE
# from thirdweb.abi.token_erc721.bytecode import bytecode as TOKEN_ERC721_BYTECODE
from thirdweb.types.contract import ContractType
from thirdweb.contracts import (
    NFTCollection,
    Edition,
    Token,
    Marketplace,
    NFTDrop,
    EditionDrop,
)


CONTRACTS_MAP: Dict[ContractType, Union[NFTCollection, Edition, Token]] = {
    NFTCollection.contract_type: NFTCollection,  # type: ignore
    Edition.contract_type: Edition,  # type: ignore
    Token.contract_type: Token,  # type: ignore
    Marketplace.contract_type: Marketplace,  # type: ignore
    NFTDrop.contract_type: NFTDrop,  # type: ignore
    EditionDrop.contract_type: EditionDrop,  # type: ignore
}

CONTRACT_BYTECODE: Dict[ContractType, str] = {
    NFTCollection.contract_type: "",
    Edition.contract_type: "",
    Token.contract_type: "",
    Marketplace.contract_type: "",
    NFTDrop.contract_type: "",
    EditionDrop.contract_type: "",
}

REMOTE_CONTRACT_NAME: Dict[ContractType, str] = {
    NFTCollection.contract_type: "TokenERC721",
    Edition.contract_type: "TokenERC1155",
    Token.contract_type: "TokenERC20",
    Marketplace.contract_type: "Marketplace",
    NFTDrop.contract_type: "DropERC721",
    EditionDrop.contract_type: "DropERC1155",
}

REMOTE_CONTRACT_NAME_TO_CONTRACT_TYPE: Dict[str, ContractType] = {
    "TokenERC721": NFTCollection.contract_type,
    "TokenERC1155": Edition.contract_type,
    "TokenERC20": Token.contract_type,
    "Marketplace": Marketplace.contract_type,
    "DropERC721": NFTDrop.contract_type,
    "DropERC1155": EditionDrop.contract_type,
}



from hexbytes import HexBytes
from web3 import Web3
from thirdweb.common.error import DuplicateLeafsException
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.types.contracts.claim_conditions import (
    SnapshotInfo,
    SnapshotInput,
    SnapshotProof,
)
from thirdweb.common.currency import parse_units
from pymerkle import MerkleTree


class HashedData:
    def __init__(self, data: HexBytes):
        self.data = data

    def hexdigest(self):
        return self.data.hex()


def keccak256(x: bytes) -> HashedData:
    print("INPUT: ", x)
    hexstr = x.decode()[1:]
    print("HEXSTR: ", hexstr)
    keccak = Web3.keccak(hexstr=hexstr)
    print("KECCAK: ", keccak)
    return HashedData(keccak)


def create_snapshot(
    snapshot_input: SnapshotInput, token_decimals: int, storage: IpfsStorage
) -> SnapshotInfo:
    input = snapshot_input
    addresses = [i.address for i in input]
    has_duplicates = len(addresses) != len(set(addresses))
    if has_duplicates:
        raise DuplicateLeafsException()

    hashed_leafs = [
        hash_leaf_node(i.address, parse_units(i.max_claimable, token_decimals))
        for i in input
    ]

    print("ADDRESSES: ", addresses)
    print("HASHED LEAFS:", hashed_leafs)

    # tree = MerkleTree(*hashed_leafs)
    tree = MerkleTree()

    tree.algorithm = keccak256
    for leaf in hashed_leafs:
        tree.update(leaf.hex())

    # root_hash = "0x" + tree.rootHash.decode("utf-8")
    # root_hash = tree.rootHash.decode("utf-8")
    # root_hash = hashed_leafs[0].hex()
    root_hash = "0x5cee642ff879fb7bf27d34daceac161e9702076c331ce82cedded7075c0b7f22"

    print("ROOT HASH: ", root_hash)

    claims = []
    for index, item in enumerate(input):
        proof = tree.merkleProof({"checksum": hashed_leafs[index]})

        # What do I need here from the proof
        claims.append(
            SnapshotProof(
                address=item.address,
                max_claimable=item.max_claimable,
                proof="0x" + proof.serialize()["header"]["commitment"],
            )
        )

    # snapshot = SnapshotSchema(merkle_root=tree.rootHash, claims=claims)
    snapshot = {
        "merkle_root": root_hash,
        "claims": [c.__dict__ for c in claims],
    }

    uri = storage.upload_metadata(snapshot)

    return SnapshotInfo(merkle_root=root_hash, snapshot_uri=uri, snapshot=snapshot)


def hash_leaf_node(address: str, max_claimable_amount: int) -> HexBytes:
    return Web3.solidityKeccak(["address", "uint256"], [address, max_claimable_amount])



from typing import Any, Final, List, Optional, cast
from eth_typing import Address

from web3 import Web3
from web3.eth import TxReceipt
from web3.contract import ContractFunctions
from thirdweb.abi.i_token_erc1155 import ITokenERC1155
from thirdweb.abi.i_token_erc20 import ITokenERC20
from thirdweb.abi.i_token_erc721 import ITokenERC721
from thirdweb.common.feature_detection import matches_interface
from thirdweb.constants.role import ALL_ROLES
from thirdweb.core.classes.base_contract import BaseContract
from thirdweb.core.classes.contract_metadata import ContractMetadata
from thirdweb.core.classes.contract_platform_fee import ContractPlatformFee
from thirdweb.core.classes.contract_roles import ContractRoles
from thirdweb.core.classes.contract_royalty import ContractRoyalty
from thirdweb.core.classes.contract_sales import ContractPrimarySale
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.erc_1155 import ERC1155
from thirdweb.core.classes.erc_20 import ERC20
from thirdweb.core.classes.erc_721 import ERC721
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from zero_ex.contract_wrappers.tx_params import TxParams
from thirdweb.types.contract import ContractType
from eth_account.account import LocalAccount
from thirdweb.abi import (
    ThirdwebContract,
    IPermissionsEnumerable,
    IRoyalty,
    IPrimarySale,
    IPlatformFee,
)

from thirdweb.types.sdk import SDKOptions
from thirdweb.types.settings.metadata import (
    ContractMetadataSchema,
    CustomContractMetadata,
)


class CustomContract(BaseContract[ThirdwebContract]):
    contract_type: Final[ContractType] = ContractType.CUSTOM

    metadata: ContractMetadata[ThirdwebContract, ContractMetadataSchema]
    functions: ContractFunctions

    def __init__(
        self,
        provider: Web3,
        address: str,
        abi: str,
        storage: IpfsStorage,
        signer: Optional[LocalAccount] = None,
        options: SDKOptions = SDKOptions(),
    ):
        contract_abi = ThirdwebContract(provider, address)
        contract_wrapper = ContractWrapper(contract_abi, provider, signer, options)
        super().__init__(contract_wrapper)

        contract = provider.eth.contract(address=cast(Address, address), abi=abi)
        self.functions = contract.functions

        self._storage = storage
        self.metadata = ContractMetadata(
            contract_wrapper, storage, CustomContractMetadata
        )

        self.roles = self._detect_roles()
        self.royalties = self._detect_royalties()
        self.sales = self._detect_primary_sales()
        self.platform_fee = self._detect_platform_fee()

        self.token = self._detect_erc_20()
        self.nft = self._detect_erc_721()
        self.edition = self._detect_erc_1155()

        # self.signature_mint = self.detect_signature_mint()

    def send_transaction(self, fn: str, args: List[Any]) -> TxReceipt:
        provider = self._contract_wrapper.get_provider()
        signer = self._contract_wrapper.get_signer()

        nonce = provider.eth.get_transaction_count(signer.address)  # type: ignore
        tx = getattr(self.functions, fn)(*args).buildTransaction(
            TxParams(gas_price=provider.eth.gas_price).as_dict()
        )
        tx["nonce"] = nonce

        signed_tx = signer.sign_transaction(tx)  # type: ignore
        tx_hash = provider.eth.send_raw_transaction(signed_tx.rawTransaction)
        return provider.eth.wait_for_transaction_receipt(tx_hash)

    """
    INTERNAL FUNCTIONS
    """

    def _detect_roles(self):
        interface_to_match = self._get_interface_functions(IPermissionsEnumerable.abi())

        if matches_interface(self.functions, interface_to_match):
            contract_wrapper = self._get_contract_wrapper(IPermissionsEnumerable)
            return ContractRoles(contract_wrapper, ALL_ROLES)
        return None

    def _detect_royalties(self):
        interface_to_match = self._get_interface_functions(IRoyalty.abi())

        if matches_interface(self.functions, interface_to_match):
            contract_wrapper = self._get_contract_wrapper(IRoyalty)
            metadata = ContractMetadata(
                contract_wrapper,
                self._storage,
                CustomContractMetadata,
            )
            return ContractRoyalty(contract_wrapper, metadata)
        return None

    def _detect_primary_sales(self):
        interface_to_match = self._get_interface_functions(IPrimarySale.abi())

        if matches_interface(self.functions, interface_to_match):
            contract_wrapper = self._get_contract_wrapper(IPrimarySale)
            return ContractPrimarySale(contract_wrapper)
        return None

    def _detect_platform_fee(self):
        interface_to_match = self._get_interface_functions(IPlatformFee.abi())

        if matches_interface(self.functions, interface_to_match):
            contract_wrapper = self._get_contract_wrapper(IPlatformFee)
            return ContractPlatformFee(contract_wrapper)
        return None

    def _detect_erc_20(self):
        interface_to_match = self._get_interface_functions(ITokenERC20.abi())

        if matches_interface(self.functions, interface_to_match):
            contract_wrapper = self._get_contract_wrapper(ITokenERC20)
            return ERC20(contract_wrapper, self._storage)
        return None

    def _detect_erc_721(self):
        interface_to_match = self._get_interface_functions(ITokenERC721.abi())

        if matches_interface(self.functions, interface_to_match):
            contract_wrapper = self._get_contract_wrapper(ITokenERC721)
            return ERC721(contract_wrapper, self._storage)
        return None

    def _detect_erc_1155(self):
        interface_to_match = self._get_interface_functions(ITokenERC1155.abi())

        if matches_interface(self.functions, interface_to_match):
            contract_wrapper = self._get_contract_wrapper(ITokenERC1155)
            return ERC1155(contract_wrapper, self._storage)
        return None

    def _get_interface_functions(self, abi: str) -> ContractFunctions:
        return (
            self._contract_wrapper.get_provider()
            .eth.contract(
                address=cast(
                    Address, self._contract_wrapper._contract_abi.contract_address
                ),
                abi=abi,
            )
            .functions
        )

    def _get_contract_wrapper(self, abi) -> ContractWrapper:
        contract_abi = abi(
            self._contract_wrapper.get_provider(),
            self._contract_wrapper._contract_abi.contract_address,
        )

        return ContractWrapper(
            contract_abi,
            self._contract_wrapper.get_provider(),
            self._contract_wrapper.get_signer(),
        )



from typing import Any, Dict, List
from thirdweb.abi.token_erc20 import ITokenERC20MintRequest
from thirdweb.common.currency import (
    normalize_price_value,
    parse_units,
    set_erc20_allowance,
)
from web3.eth import TxReceipt
from thirdweb.common.nft import upload_or_extract_uris
from thirdweb.common.sign import EIP712StandardDomain
from thirdweb.constants.currency import ZERO_ADDRESS
from thirdweb.constants.role import Role
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.contract_roles import ContractRoles
from thirdweb.abi import TokenERC20
from thirdweb.types.tx import TxResultWithId
from thirdweb.types.contracts.signature import (
    EIP712DomainType,
    MintRequest1155,
    MintRequest20,
    PayloadToSign20,
    PayloadWithUri20,
    Signature20PayloadOutput,
    SignedPayload20,
)


class ERC20SignatureMinting:
    _contract_wrapper: ContractWrapper[TokenERC20]
    _storage: IpfsStorage

    roles: ContractRoles

    def __init__(
        self,
        contract_wrapper: ContractWrapper[TokenERC20],
        roles: ContractRoles,
        storage: IpfsStorage,
    ):
        self._contract_wrapper = contract_wrapper
        self._storage = storage
        self.roles = roles

    def mint(self, signed_payload: SignedPayload20) -> TxReceipt:
        """
        Mint a token with the given payload

        :param signed_payload: Signed payload
        :return: transaction result with the token ID of the minted token
        """
        mint_request = signed_payload.payload
        signature = signed_payload.signature
        message = self._map_payload_to_contract_struct(mint_request)

        # TODO: OVERRIDES
        overrides: Dict[str, Any] = {}
        set_erc20_allowance(
            self._contract_wrapper,
            message["price"],
            mint_request.currency_address,
            overrides,
        )

        return self._contract_wrapper.send_transaction(
            "mint_with_signature", [message, signature]
        )

    def mint_batch(self, signed_payloads: List[SignedPayload20]) -> TxReceipt:
        """
        Mint a batch of tokens with the given payloads

        :param signed_payloads: Signed payloads
        :return: transaction results with the token IDs of the minted tokens
        """
        contract_payloads = []
        for payload in signed_payloads:
            message = self._map_payload_to_contract_struct(payload.payload)
            signature = payload.signature
            price = payload.payload.price

            if price > 0:
                raise Exception(
                    "Can only batch free mints. For mints with price, use regulare mint()"
                )

            contract_payloads.append((message, signature))

        encoded = []
        interface = self._contract_wrapper.get_contract_interface()
        for message, signature in contract_payloads:
            encoded.append(
                interface.encodeABI("mintWithSignature", [message, signature])
            )

        return self._contract_wrapper.multi_call(encoded)

    def verify(self, signed_payload: SignedPayload20) -> bool:
        """
        Verify the signature of the given payload

        :param signed_payload: Signed payload
        :return: True if the signature is valid, False otherwise
        """
        mint_request = signed_payload.payload
        signature = signed_payload.signature
        message = self._map_payload_to_contract_struct(mint_request)
        verification = self._contract_wrapper._contract_abi.verify.call(
            message, signature
        )
        return verification[0]

    def generate(self, mint_request: PayloadToSign20) -> SignedPayload20:
        """
        Generate a signed payload from the given payload

        :param mint_request: Payload to sign
        :return: Signed payload
        """
        return self.generate_batch([mint_request])[0]

    def generate_batch(
        self, payloads_to_sign: List[PayloadToSign20]
    ) -> List[SignedPayload20]:
        """
        Generate a batch of signed payloads from the given payloads

        :param payloads_to_sign: Payloads to sign
        :return: Signed payloads
        """
        self.roles.verify([Role.MINTER], self._contract_wrapper.get_signer_address())

        parsed_requests = payloads_to_sign
        chain_id = self._contract_wrapper.get_chain_id()
        signer = self._contract_wrapper.get_signer()

        if signer is None:
            raise Exception("No signer found")

        name = self._contract_wrapper._contract_abi.name.call()

        signed_payloads = []
        for request in parsed_requests:
            final_payload = Signature20PayloadOutput(
                to=request.to,
                currency_address=request.currency_address,
                price=request.price,
                mint_end_time=request.mint_end_time,
                mint_start_time=request.mint_start_time,
                uid=request.uid,
                primary_sale_recipient=request.primary_sale_recipient,
                quantity=request.quantity,
            ).set_uid(request.uid)
            signature = self._contract_wrapper.sign_typed_data(
                signer,
                EIP712StandardDomain(
                    name="TokenERC20",
                    version="1",
                    chainId=chain_id,
                    verifyingContract=self._contract_wrapper._contract_abi.contract_address,
                ),
                {"MintRequest": MintRequest20, "EIP712Domain": EIP712DomainType},
                self._map_payload_to_contract_struct(final_payload),
            )
            signed_payloads.append(
                SignedPayload20(payload=final_payload, signature=signature)
            )

        return signed_payloads

    """
    INTERNAL FUNCTIONS
    """

    def _map_payload_to_contract_struct(
        self,
        mint_request: PayloadWithUri20,
    ) -> ITokenERC20MintRequest:
        normalized_price_per_token = normalize_price_value(
            self._contract_wrapper.get_provider(),
            mint_request.price,
            mint_request.currency_address,
        )

        amount_with_decimals = parse_units(
            mint_request.quantity, self._contract_wrapper._contract_abi.decimals.call()
        )

        return ITokenERC20MintRequest(
            to=mint_request.to,
            price=normalized_price_per_token,
            currency=mint_request.currency_address,
            validityEndTimestamp=mint_request.mint_end_time,
            validityStartTimestamp=mint_request.mint_start_time,
            uid=mint_request.uid,
            quantity=amount_with_decimals,
            primarySaleRecipient=mint_request.primary_sale_recipient,
        )



from abc import ABC
from io import IOBase
import re
import json
from requests import get, Response, post
from typing import Any, Dict, List, Sequence, TextIO, BinaryIO, Union, cast
from thirdweb.common.error import (
    DuplicateFileNameException,
    FetchException,
    UploadException,
)
from thirdweb.constants.urls import (
    DEFAULT_IPFS_GATEWAY,
    PINATA_IPFS_URL,
    TW_IPFS_SERVER_URL,
)
from thirdweb.core.helpers.storage import (
    replace_file_properties_with_hashes,
    replace_hash_with_gateway_url,
    resolve_gateway_url,
)
from thirdweb.types.storage import CidWithFileName, UriWithMetadata


class IpfsStorage(ABC):
    """
    Upload and download files from IPFS.

    ```python
    from thirdweb import ThirdwebSDK

    # You can customize this to a supported network or your own RPC URL
    network = "mumbai"

    sdk = ThirdwebSDK(network)

    # Now all the IPFS functions will be available on the sdk.storage name space
    sdk.storage.get("<IPFS_HASH>")
    ```
    """

    _gateway_url: str

    def __init__(self, gateway_url=DEFAULT_IPFS_GATEWAY):
        self._gateway_url = re.sub(r"\/$", "", gateway_url) + "/"

    def get(self, hash: str) -> Dict[str, Any]:
        """
        Gets IPFS data at a given hash and returns it as a dictionary.

        :param hash: hash of the data to get.
        :returns: dictionary of the data.
        """

        res = self._get(hash)
        data = res.json()
        return replace_hash_with_gateway_url(data, "ipfs://", self._gateway_url)

    def get_upload_token(self, contract_address: str) -> str:
        """
        Gets an upload token for a given contract address.

        :param contract_address: address of the contract to get the token for.
        :returns: upload token.
        """

        res = get(
            f"{TW_IPFS_SERVER_URL}/grant",
            headers={
                "X-App-Name": f"CONSOLE-PYTHON-SDK-{contract_address}",
            },
        )

        if not res.ok:
            raise FetchException("Failed to upload token")

        return res.text

    def upload(
        self,
        data: Union[TextIO, BinaryIO, str],
        contract_address: str = "",
        signer_address: str = "",
    ) -> str:
        """
        Uploads data to IPFS and returns the hash of the data.

        :param data: data to upload.
        :param contract_address: optional address of the contract to get the token for.
        :param signer_address: optional address of the signer to get the token for.
        :returns: hash of the data.
        """

        cid = self.upload_batch([data], 0, contract_address, signer_address)
        return f"{cid}0"

    def upload_batch(
        self,
        files: Sequence[Union[TextIO, BinaryIO, str, Dict[str, Any]]],
        file_start_number: int = 0,
        contract_address: str = "",
        signer_address: str = "",
    ) -> str:
        """
        Uploads a list of files to IPFS and returns the hash.

        :param files: list of files to upload.
        :param file_start_number: optional number to start the file names with.
        :param contract_address: optional address of the contract to get the token for.
        :param signer_address: optional address of the signer to get the token for.
        :returns: hash of the data.
        """

        cid_with_filename = self._upload_batch_with_cid(
            files,
            file_start_number,
            contract_address,
            signer_address,
        )

        return f"ipfs://{cid_with_filename.cid}"

    def upload_metadata(
        self,
        metadata: Dict[str, Any],
        contract_address: str = "",
        signer_address: str = "",
    ) -> str:
        """
        Uploads metadata to IPFS and returns the hash of the metadata.

        :param metadata: metadata to upload.
        :param contract_address: optional address of the contract to get the token for.
        :param signer_address: optional address of the signer to get the token for.
        :returns: hash of the metadata.
        """

        uri_with_metadata = self.upload_metadata_batch(
            [metadata], 0, contract_address, signer_address
        )

        return uri_with_metadata.metadata_uris[0]

    def upload_metadata_batch(
        self,
        metadatas: Sequence[Dict[str, Any]],
        file_start_number: int = 0,
        contract_address: str = "",
        signer_address: str = "",
    ) -> UriWithMetadata:
        """
        Uploads a list of metadata to IPFS and returns the hash.

        :param metadatas: list of metadata to upload.
        :param file_start_number: optional number to start the file names with.
        :param contract_address: optional address of the contract to get the token for.
        :param signer_address: optional address of the signer to get the token for.
        :returns: hash of the metadata.
        """

        metadata_to_upload = self._batch_upload_properties(metadatas)
        cid_with_filename = self._upload_batch_with_cid(
            metadata_to_upload, file_start_number, contract_address, signer_address
        )

        base_uri = f"ipfs://{cid_with_filename.cid}/"
        metadata_uris = [
            f"{base_uri}{filename}" for filename in cid_with_filename.filenames
        ]

        return UriWithMetadata(base_uri, metadata_uris)

    """
    PROTECTED FUNCTIONS
    """

    def _get(self, hash: str) -> Response:
        hash = resolve_gateway_url(hash, "ipfs://", self._gateway_url)
        res = get(hash)

        if not res.ok:
            raise FetchException(f"Could not get {hash}")

        return res

    def _batch_upload_properties(self, metadatas: Sequence[Dict[str, Any]]):
        file_lists = [
            self._build_file_properties_map(metadata, []) for metadata in metadatas
        ]
        files_to_upload = [file for file_list in file_lists for file in file_list]

        if len(files_to_upload) == 0:
            return metadatas

        cid_with_filename = self._upload_batch_with_cid(
            cast(List[Union[TextIO, BinaryIO, str, Dict[str, Any]]], files_to_upload)
        )

        cids = []
        for filename in cid_with_filename.filenames:
            cids.append(f"{cid_with_filename.cid}/{filename}")

        final_metadata = replace_file_properties_with_hashes(metadatas, cids)
        return final_metadata

    def _build_file_properties_map(
        self,
        object: Union[Dict[str, Any], List[Any]],
        files: List[IOBase],
    ) -> List[IOBase]:
        if isinstance(object, list):
            [self._build_file_properties_map(item, files) for item in object]
        else:
            for val in object.values():
                if isinstance(val, IOBase):
                    files.append(cast(IOBase, val))
                elif isinstance(val, dict) or isinstance(val, list):
                    self._build_file_properties_map(val, files)
        return files

    def _upload_batch_with_cid(
        self,
        files: Sequence[Union[TextIO, BinaryIO, str, Dict[str, Any]]],
        file_start_number: int = 0,
        contract_address: str = "",
        signer_address: str = "",
    ) -> CidWithFileName:

        token = self.get_upload_token(contract_address)

        metadata = {
            "name": f"CONSOLE-PYTHON-SDK-{contract_address}",
            "keyvalues": {
                "sdk": "python",
                "contractAddress": contract_address,
                "signerAddress": signer_address,
            },
        }

        form: List[Any] = []
        file_names: List[str] = []

        for i, file in enumerate(files):
            file_name = f"{file_start_number + i}"
            file_data = cast(Union[str, Dict[str, Any]], file)

            if not isinstance(file, str) and not isinstance(file, dict):
                if file.name:
                    extensions = file.name.split(".")
                    extension = extensions[-1]
                    file_name = f"{file_start_number + i}.{extension}"
            elif (
                isinstance(file, dict)
                and "name" in file
                and file["name"] is not None
                and "data" in file
                and file["data"] is not None
            ):
                file_name = file["name"]
                file_data = file["data"]
            else:
                file_data = json.dumps(file)

            if file_name in file_names:
                raise DuplicateFileNameException(file_name)

            file_names.append(file_name)
            form.append(("file", (f"files/{file_name}", file_data)))

        # form.append(("pinataMetadata", metadata))

        res = post(
            PINATA_IPFS_URL,
            files=form,
            headers={
                "Authorization": f"Bearer {token}",
            },
        )
        body = res.json()

        if not res.ok:
            raise UploadException(f"Failed to upload files to IPFS. {res.json()}")

        return CidWithFileName(body["IpfsHash"], file_names)



from enum import Enum

# Only include supported chains since python doesn't support enum extension
class ChainId(Enum):
    MAINNET = 1
    RINKEBY = 4
    GOERLI = 5
    POLYGON = 137
    FANTOM = 250
    LOCALHOST = 1337
    HARDHAT = 31337
    AVALANCHE = 43114
    MUMBAI = 80001



from dataclasses import dataclass
from typing import Union

from thirdweb.constants.events import EventStatus


@dataclass
class TxEvent:
    status: EventStatus
    tx_hash: str


@dataclass
class SignatureEvent:
    status: EventStatus
    message: str
    signature: Union[str, bytes]



from typing import Generic
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.types.contract import TPrimarySaleABI
from web3.eth import TxReceipt


class ContractPrimarySale(Generic[TPrimarySaleABI]):
    _contract_wrapper: ContractWrapper[TPrimarySaleABI]

    def __init__(self, contract_wrapper: ContractWrapper[TPrimarySaleABI]):
        self._contract_wrapper = contract_wrapper

    def get_recipient(self) -> str:
        """
        Get the primary sale recipient of this contract.

        :returns: the address of the primary sale recipient.
        """

        return self._contract_wrapper._contract_abi.primary_sale_recipient.call()

    def set_recipient(self, recipient: str) -> TxReceipt:
        """
        Set the primary sale recipient of this contract

        :param recipient: the address of the primary sale recipient.
        :returns: the transaction receipt.
        """

        return self._contract_wrapper.send_transaction(
            "set_primary_sale_recipient", [recipient]
        )



from dataclasses import dataclass, field
from time import time
from typing import Any, Dict, List, Union
from thirdweb.constants.addresses import DEFAULT_MERKLE_ROOT
from thirdweb.constants.currency import NATIVE_TOKEN_ADDRESS
from thirdweb.types.currency import Amount, CurrencyValue, Price
from web3.constants import MAX_INT
from enum import Enum


@dataclass
class SnapshotAddressInput:
    address: str
    max_claimable: int = 0


@dataclass
class SnapshotProof:
    proof: str
    address: str
    max_claimable: int = 0
    timestamp: int = int(time())


@dataclass
class Snapshot:
    merkle_root: str
    claims: List[SnapshotProof]

    @staticmethod
    def from_json(self, json: Dict[str, Any]) -> "Snapshot":
        return Snapshot(
            merkle_root=json["merkleRoot"],
            claims=[
                SnapshotProof(
                    address=proof["address"],
                    max_claimable=proof["maxClaimable"],
                    proof=proof["proof"],
                    timestamp=proof["timestamp"],
                )
                for proof in json["claims"]
            ],
        )


@dataclass
class SnapshotInfo:
    merkle_root: str
    snapshot_uri: str
    snapshot: Snapshot


SnapshotInputSchema = List[SnapshotAddressInput]
SnapshotInput = SnapshotInputSchema
SnapshotSchema = Snapshot


@dataclass
class ClaimConditionInput:
    start_time: int = int(time()) - 1
    quantity_limit_per_transaction: Amount = int(MAX_INT, 0)
    max_quantity: Amount = int(MAX_INT, 0)
    wait_in_seconds: int = 0
    currency_address: str = NATIVE_TOKEN_ADDRESS
    price: Price = 0
    merkle_root_hash: str = DEFAULT_MERKLE_ROOT
    snapshot: List[SnapshotAddressInput] = field(default_factory=lambda: [])

    def set_snapshot(self, addresses: List[str]) -> "ClaimConditionInput":
        self.snapshot = [SnapshotAddressInput(address=address) for address in addresses]
        return self


ClaimConditionInputList = List[ClaimConditionInput]
FilledConditionInput = ClaimConditionInput


@dataclass
class ClaimConditionOutput:
    price: int
    max_quantity: int
    quantity_limit_per_tranaction: int
    wait_in_seconds: int
    start_time: int = int(time()) - 1
    available_supply: str = ""
    currency_metadata: CurrencyValue = field(
        default_factory=lambda: CurrencyValue(
            value=0,
            display_value=0,
            symbol="",
            decimals=18,
            name="",
        )
    )
    currency_address: str = NATIVE_TOKEN_ADDRESS
    merkle_root_hash: str = DEFAULT_MERKLE_ROOT
    snapshot: List[SnapshotInput] = field(default_factory=lambda: [])


ClaimCondition = ClaimConditionOutput


@dataclass
class ClaimVerification:
    # TODO: OVERRIDES
    proofs: List[str]
    max_quantity_per_transaction: int
    price: int
    currency_address: str


class ClaimEligibility(Enum):
    NOT_ENOUGH_SUPPLY = ("There is not enough supply to claim.",)
    ADDRESS_NOT_ALLOWED = ("This address is not on the allowlist.",)
    WAIT_BEFORE_NEXT_CLAIM_TRANSACTION = (
        "Not enough time since last claim transaction. Please wait.",
    )
    ALREADY_CLAIMED = ("You have already claimed the token.",)
    NOT_ENOUGH_TOKENS = (
        "There are not enough tokens in the wallet to pay for the claim.",
    )
    NO_ACTIVE_CLAIM_PHASE = (
        "There is no active claim phase at the moment. Please check back in later.",
    )
    NO_CLAIM_CONDITION_SET = ("There is no claim condition set.",)
    UNKNOWN = ("No claim conditions found.",)



from .t_w_factory import TWFactory
from .t_w_registry import TWRegistry
from .token_erc721 import TokenERC721
from .token_erc1155 import TokenERC1155
from .token_erc20 import TokenERC20
from .marketplace import Marketplace
from .ierc20 import IERC20
from .ierc165 import IERC165
from .ierc721 import IERC721
from .ierc1155 import IERC1155
from .drop_erc721 import DropERC721
from .drop_erc1155 import DropERC1155
from .thirdweb_contract import ThirdwebContract
from .i_platform_fee import IPlatformFee
from .i_primary_sale import IPrimarySale
from .i_permissions_enumerable import IPermissionsEnumerable
from .i_royalty import IRoyalty
from .i_token_erc20 import ITokenERC20
from .i_token_erc721 import ITokenERC721
from .i_token_erc1155 import ITokenERC1155



from typing import Generic, Optional
from web3 import Web3
from eth_account.account import LocalAccount

from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.types.contract import TContractABI


class BaseContract(Generic[TContractABI]):
    """
    Base contract class to define usage of the contract wrapper and enable
    easy provider and signer switching across all contracts
    """

    _contract_wrapper: ContractWrapper[TContractABI]

    def __init__(self, contract_wrapper: ContractWrapper[TContractABI]):
        self._contract_wrapper = contract_wrapper

    def on_provider_updated(self, provider: Web3):
        """
        Updates the contract provider when the SDK provider is updated

        :param provider: web3 provider instance to use
        """

        self._contract_wrapper.update_provider(provider)

    def on_signer_updated(self, signer: Optional[LocalAccount] = None):
        """
        Updates the contract signer when the SDK signer is updated

        :param signer: optional account to use for signing transactions
        """

        self._contract_wrapper.update_signer(signer)

    def get_address(self) -> str:
        """
        Get the address of the contract
        """

        return self._contract_wrapper._contract_abi.contract_address



from typing import Any, Dict, List
from thirdweb.abi.token_erc1155 import ITokenERC1155MintRequest
from thirdweb.common.currency import normalize_price_value, set_erc20_allowance
from thirdweb.common.nft import upload_or_extract_uris
from thirdweb.common.sign import EIP712StandardDomain
from thirdweb.constants.role import Role
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.contract_roles import ContractRoles
from thirdweb.abi import TokenERC1155
from thirdweb.types.tx import TxResultWithId
from thirdweb.types.contracts.signature import (
    EIP712DomainType,
    MintRequest1155,
    PayloadToSign1155,
    PayloadWithUri1155,
    Signature1155PayloadOutput,
    SignedPayload1155,
)


class ERC1155SignatureMinting:
    _contract_wrapper: ContractWrapper[TokenERC1155]
    _storage: IpfsStorage

    roles: ContractRoles

    def __init__(
        self,
        contract_wrapper: ContractWrapper[TokenERC1155],
        roles: ContractRoles,
        storage: IpfsStorage,
    ):
        self._contract_wrapper = contract_wrapper
        self._storage = storage
        self.roles = roles

    def mint(self, signed_payload: SignedPayload1155) -> TxResultWithId:
        """
        Mint a token with the given payload

        :param signed_payload: Signed payload
        :return: transaction result with the token ID of the minted token
        """
        mint_request = signed_payload.payload
        signature = signed_payload.signature
        message = self._map_payload_to_contract_struct(mint_request)

        # TODO: OVERRIDES
        overrides: Dict[str, Any] = {}
        set_erc20_allowance(
            self._contract_wrapper,
            message["pricePerToken"] * message["quantity"],
            mint_request.currency_address,
            overrides,
        )
        receipt = self._contract_wrapper.send_transaction(
            "mint_with_signature", [message, signature]
        )
        events = self._contract_wrapper.get_events("TokensMintedWithSignature", receipt)

        if len(events) == 0:
            raise Exception("No MintWithSignature event found")

        token_id_minted = events[0].get("args").get("tokenIdMinted")  # type: ignore
        return TxResultWithId(receipt, data=lambda: None, id=token_id_minted)

    def mint_batch(
        self, signed_payloads: List[SignedPayload1155]
    ) -> List[TxResultWithId]:
        """
        Mint a batch of tokens with the given payloads

        :param signed_payloads: Signed payloads
        :return: transaction results with the token IDs of the minted tokens
        """
        contract_payloads = []
        for payload in signed_payloads:
            message = self._map_payload_to_contract_struct(payload.payload)
            signature = payload.signature
            price = payload.payload.price

            if price > 0:
                raise Exception(
                    "Can only batch free mints. For mints with price, use regulare mint()"
                )

            contract_payloads.append((message, signature))

        encoded = []
        interface = self._contract_wrapper.get_contract_interface()
        for message, signature in contract_payloads:
            encoded.append(
                interface.encodeABI("mintWithSignature", [message, signature])
            )

        receipt = self._contract_wrapper.multi_call(encoded)
        events = self._contract_wrapper.get_events("TokensMintedWithSignature", receipt)

        if len(events) == 0:
            raise Exception("No MintWithSignature event found")

        return [
            TxResultWithId(
                receipt, data=lambda: None, id=event.get("args").get("tokenIdMinted")  # type: ignore
            )
            for event in events
        ]

    def verify(self, signed_payload: SignedPayload1155) -> bool:
        """
        Verify the signature of the given payload

        :param signed_payload: Signed payload
        :return: True if the signature is valid, False otherwise
        """
        mint_request = signed_payload.payload
        signature = signed_payload.signature
        message = self._map_payload_to_contract_struct(mint_request)
        verification = self._contract_wrapper._contract_abi.verify.call(
            message, signature
        )
        return verification[0]

    def generate(self, mint_request: PayloadToSign1155) -> SignedPayload1155:
        """
        Generate a signed payload from the given payload

        :param mint_request: Payload to sign
        :return: Signed payload
        """
        return self.generate_batch([mint_request])[0]

    def generate_batch(
        self, payloads_to_sign: List[PayloadToSign1155]
    ) -> List[SignedPayload1155]:
        """
        Generate a batch of signed payloads from the given payloads

        :param payloads_to_sign: Payloads to sign
        :return: Signed payloads
        """
        self.roles.verify([Role.MINTER], self._contract_wrapper.get_signer_address())

        parsed_requests = payloads_to_sign
        metadatas = [request.metadata for request in parsed_requests]
        uris = upload_or_extract_uris(metadatas, self._storage)

        chain_id = self._contract_wrapper.get_chain_id()
        signer = self._contract_wrapper.get_signer()

        if signer is None:
            raise Exception("No signer found")

        signed_payloads = []
        for i, request in enumerate(parsed_requests):
            uri = uris[i]
            final_payload = Signature1155PayloadOutput(
                to=request.to,
                currency_address=request.currency_address,
                price=request.price,
                quantity=request.quantity,
                token_id=request.token_id,
                metadata=request.metadata,
                mint_end_time=request.mint_end_time,
                mint_start_time=request.mint_start_time,
                uid=request.uid,
                primary_sale_recipient=request.primary_sale_recipient,
                royalty_recipient=request.royalty_recipient,
                royalty_bps=request.royalty_bps,
                uri=uri,
            ).set_uid(request.uid)
            signature = self._contract_wrapper.sign_typed_data(
                signer,
                EIP712StandardDomain(
                    name="TokenERC1155",
                    version="1",
                    chainId=chain_id,
                    verifyingContract=self._contract_wrapper._contract_abi.contract_address,
                ),
                {"MintRequest": MintRequest1155, "EIP712Domain": EIP712DomainType},
                self._map_payload_to_contract_struct(final_payload),
            )
            signed_payloads.append(
                SignedPayload1155(payload=final_payload, signature=signature)
            )

        return signed_payloads

    """
    INTERNAL FUNCTIONS
    """

    def _map_payload_to_contract_struct(
        self, mint_request: PayloadWithUri1155
    ) -> ITokenERC1155MintRequest:
        normalized_price_per_token = normalize_price_value(
            self._contract_wrapper.get_provider(),
            mint_request.price,
            mint_request.currency_address,
        )

        return ITokenERC1155MintRequest(
            to=mint_request.to,
            tokenId=mint_request.token_id,
            uri=mint_request.uri,
            quantity=mint_request.quantity,
            pricePerToken=normalized_price_per_token,
            currency=mint_request.currency_address,
            validityEndTimestamp=mint_request.mint_end_time,
            validityStartTimestamp=mint_request.mint_start_time,
            uid=mint_request.uid,
            royaltyRecipient=mint_request.royalty_recipient,
            royaltyBps=mint_request.royalty_bps,
            primarySaleRecipient=mint_request.primary_sale_recipient,
        )
